//setarea real-time-clock (RTC)
//se porneste un alt timer intr-o operatie asincrona folosind un quartz de 32.768kHz
#include <stdint.h>
#include <avr/interrupt.h>
#include <util/delay.h>

volatile uint8_t RTC_h, RTC_m, RTC_s, RTC_secondchanged,
	RTC_minutechanged;


void delay(uint16_t millisec) 
{
	while (millisec--) {
		_delay_ms(1);
	}
}

void RTC_init(void)
{
	uint8_t sreg;
	//salvez registrul de stare si dezactivez intreruperile
	sreg=SREG;
	cli();
	//setez un delay de 1 sec pentru ca oscilatorul sa aiba timp sa se stabilizeze
	delay(1000);
    //dezactivez intreruperile noului timer prin setarea OCIE2 si TOIE2
	TIMSK &= ~((1<<OCIE2)|(1<<TOIE2));
	//selectez sursa ceasului prin setarea AS2
	//schimb valorile ptr reg TCNT2  (pe 0), OCR2 si TCCR2
	ASSR = (1<<AS2);
	TCNT2 = 0;
	//selectez prescaler 32.768/128= 1 sec intre fiecare overflow
	TCCR2 |= (1<<CS22) | (1<<CS20);            
	//schimb operatia in asincrona
	while( (ASSR & (1<<TCN2UB)) | (ASSR & (1<<OCR2UB)) |(ASSR & (1<<TCR2UB)) ); 
	
	//resetez flag-urile de intreruperi ptr timer
	TIFR |= ((1<<OCF2)|(1<<TOV2));
	
	//activez intreruperea ptr overflow-ul timerului 
	TIMSK |= (1<<TOIE2);
    
	//initializez valorile ptr ora,minut,secunda
	RTC_h=0;
	RTC_m=0;
	RTC_s=0;
	RTC_secondchanged = 1;
	RTC_minutechanged = 1;
	
	//setez registrul de stare cu ce am retinut la inceput
	SREG=sreg;
}

//functie care seteaza ora, minutul ,secunda
void RTC_set(uint8_t th, uint8_t tm, uint8_t ts)
{
	uint8_t sreg;
	//retin registrul de stare
	sreg=SREG;
	//dezactivez intreruperile
	cli();
	//setez ora , minutul, secunda
	RTC_h=th; 
	RTC_m=tm; 
	RTC_s=ts;
	RTC_secondchanged = 1;
	RTC_minutechanged = 1;
	//resetez registrul de stare
	SREG=sreg;
}

//semnalul de intrerupere overflow timer este apelat din sec in sec
ISR(TIMER2_OVF_vect)
{//se seteaza regulile dupa care sunt schimbate ora, minutul si secunda
	//creste sec
	RTC_secondchanged = 1;
	RTC_s++;
	//daca se ajunge la secunda 60
	if (RTC_s == 60)
	{
		RTC_s = 0;//resetez nr de secunde la 0
		RTC_m++;//cresc minutele
		RTC_minutechanged = 1;
		//daca minutele ajung la 60
		if (RTC_m > 59)
		{
			RTC_m = 0;//resetez minutele la 0
			RTC_h++;//cresc ora
			//daca ora ajunge la 24 o resetez la 0
			if (RTC_h > 23) RTC_h = 0;
		}
	}
}

